package es.uca.xmleye.presentation;

/*
  Copyright (C) 2008 Antonio García Domínguez

  This program is free software; you can redistribute it
  and/or modify it under the terms of the GNU General Public
  License as published by the Free Software Foundation;
  either version 2 of the License, or (at your option) any
  later version.

  This program is distributed in the hope that it will be
  useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
  PURPOSE. See the GNU General Public License for more
  details.

  You should have received a copy of the GNU General Public
  License along with this program; if not, write to the Free
  Software Foundation, Inc., 59 Temple Place, Suite 330,
  Boston, MA 02111-1307 USA
*/

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ExecutionException;
import java.util.Queue;
import java.util.Locale;
import java.util.ResourceBundle;

import es.uca.xmleye.ImportProcessExecutionException;
import es.uca.xmleye.ParsingException;
import es.uca.xmleye.ImportException;
import es.uca.xmleye.DocumentModel;
import es.uca.xmleye.formats.*;
import es.uca.xmleye.gui.Application;
import es.uca.xmleye.prefs.*;
import es.uca.xmleye.util.Pair;
import es.uca.xmleye.util.ValueWrapper;
import es.uca.xmleye.util.FilterProcess;
import es.uca.xmleye.util.LineDemuxTask;
import es.uca.xmleye.util.LineFilter;
import es.uca.xmleye.util.WriteToFileLineFilter;
import es.uca.xmleye.util.SaveToMemoryLineFilter;
import es.uca.xmleye.xslt.*;

/**
 * Modelo de presentación de la aplicación completa.
 */
public class WndMainPresentationModel {
  private static final ResourceBundle errRes;
  private static final ResourceBundle miscRes;

  static {
    miscRes = ResourceBundle.getBundle("es.uca.xmleye.l10n.MiscMessages",
                                       Locale.getDefault());
    errRes = ResourceBundle.getBundle("es.uca.xmleye.l10n.ErrorMessages",
                                      Locale.getDefault());
  }
  private PreferencesManager gestorPrefs; // Preferencias actuales
  private StylesheetRepository repoHojas; // Hojas disponibles

  // Documentos abiertos actualmente, según fichero origen
  private Map<File, DocumentPresentationModel> docsAbiertos;
  private RecentDocumentHistory docsRecientes; // RecentDocumentHistory de
  // documentos recientes
  private PreferenceBackedBoolean bActAutomatica; // ¿Actualización
  // automática?
  private PreprocessingStylesheet hojaPreprocOmision; // Stylesheet de
  // preprocesado por
  // omisión
  private ViewStylesheet hojaVistaOmision; // Stylesheet de visualización
  // por
  // omisión
  private final Queue<ValueWrapper<String>> mensajesTareas; // Mensajes de estado de
  private FormatDescriptorRepository formatRepository;

  // Modelo vacío de documento (para cuando no hay nada abierto)
  private final DocumentPresentationModel docVacio;

  // las tareas en
  // ejecución y
  // pendientes

  /* --- CONSTRUCTORES --- */
  public WndMainPresentationModel(
      FormatDescriptorRepository formatRepository,
      PreferencesManager gestorPrefs, StylesheetRepository repoHojas)
      throws ParsingException, TransformationException,
      StylesheetRepositoryException {

    this.formatRepository = formatRepository;
    this.gestorPrefs = gestorPrefs;
    this.docsAbiertos = new HashMap<File, DocumentPresentationModel>();
    this.repoHojas = repoHojas;

    this.docsRecientes = new RecentDocumentHistory(gestorPrefs);
    this.bActAutomatica = new PreferenceBackedBoolean(gestorPrefs,
                                                      DefaultPreferences.CLAVE_AUTOIMPORT,
                                                      DefaultPreferences.INICIAL_AUTOIMPORT);

    this.hojaPreprocOmision = repoHojas.getHojaPreprocesado(gestorPrefs
                                                            .getScalarPref(DefaultPreferences.CLAVE_PREPROCESADO_OMISION,
                                                                           DefaultPreferences.INICIAL_PREPROCESADO_OMISION));
    this.hojaVistaOmision = repoHojas.getHojaVisualizacion(gestorPrefs
                                                           .getScalarPref(DefaultPreferences.CLAVE_VISTA_OMISION,
                                                                          DefaultPreferences.INICIAL_VISTA_OMISION));

    this.mensajesTareas = new ConcurrentLinkedQueue<ValueWrapper<String>>();
    this.docVacio = new DocumentPresentationModel(this);
  }

  /* ------------------------- METODOS PUBLICOS ---------------------------- */

  /* -------- Accesores para sincronización -------- */
  public List<String> getEntradasHistorial() {
    return docsRecientes.getHistorial();
  }

  public List<String> getNombresHojasPreproc() {
    return repoHojas.getNombresHojasPreprocesado();
  }

  public List<String> getNombresHojasVista() {
    return repoHojas.getNombresHojasVisualizacion();
  }

  public boolean getActualizacionAutomatica() {
    return bActAutomatica.get();
  }

  public String getTitulo(DocumentPresentationModel docActual) {
    return "XMLEye "
        + Application.VERSION
        + ((docActual != null && docActual.getNombreDocumento() != null) ? (" - " + docActual
                                                                            .getNombreDocumento())
           : "");
  }

  public boolean getOcupado() {
    return !mensajesTareas.isEmpty();
  }

  public String getMensajeEstado() {
    synchronized (mensajesTareas) {
      if (mensajesTareas.isEmpty()) {
        return miscRes.getString("status_ready");
      } else {
        return mensajesTareas.peek().get();
      }
    }
  }

  /* --- Respuesta a eventos --- */

  // / Lanzamiento de solicitud de apertura de un fichero
  public DocumentPresentationModel solicitudAbrirDocumento(File fich)
      throws ParsingException, TransformationException, ImportException {
    final PreprocessingStylesheet hojaPreproc;
    final ViewStylesheet hojaVista;
    final ValueWrapper<String> elem;

    synchronized (this) {
      hojaPreproc = hojaPreprocOmision;
      hojaVista = hojaVistaOmision;
      elem = inicioTarea(miscRes.getString("status_opening"));
    }

    // Lo quitamos del historial por si hay un fallo
    docsRecientes.retirar(fich.getAbsolutePath());

    try {
      final FormatDescriptor formato = formatRepository
          .getFormatoPorFich(fich);

      if (formato != null && formato.getPatronOrdenImportacion() != null
          && formato.getPatronOrdenImportacion().trim().length() > 0) {

        // Invoke the import process, dumping its stdout to tmpFile
        File tmpFile = File.createTempFile("gen", null);
        finTarea(elem);

        final SaveToMemoryLineFilter savedStdout = new SaveToMemoryLineFilter();
        final SaveToMemoryLineFilter savedStderr = new SaveToMemoryLineFilter();
        FilterProcess proc = new FilterProcess(
            formato.getOrdenImportacion(fich),
            null,
            new LineDemuxTask(new LineFilter[]{
                new WriteToFileLineFilter(tmpFile),
                savedStdout}),
            new LineDemuxTask(new LineFilter[]{
                new ProgramStatusLineFilter(),
                savedStderr}));

        int codigoProc = proc.iniciar(true);
        if (codigoProc != 0) {
          throw new ImportProcessExecutionException(
              String.format(errRes.getString("err_importer_fail"), codigoProc),
              codigoProc,
              savedStdout.getSavedContents(),
              savedStderr.getSavedContents());
        }

        DocumentPresentationModel res = new DocumentPresentationModel(
            new DocumentModel(tmpFile, hojaPreproc, fich, fich
                              .getAbsoluteFile().getParent()), hojaVista,
            this);

        return res;
      } else {

        DocumentPresentationModel res = new DocumentPresentationModel(
            new DocumentModel(fich, hojaPreproc, fich, fich
                              .getAbsoluteFile().getParent()), hojaVista,
            this);

        return res;
      }

    } catch (FileNotFoundException ex) {
      throw new ImportException(ex);
    } catch (IOException ex) {
      throw new ImportException(String.format(errRes
                                              .getString("err_import_ioex"), ex.getLocalizedMessage()),
                                ex);
    } catch (InterruptedException ex) {
      throw new ImportException(ex);
    } catch (ParsingException ex) {
      throw ex;
    } catch (TransformationException ex) {
      throw ex;
    } catch (ExecutionException e) {
      throw new ImportException(e);
    } finally {
      if (elem != null) {
        finTarea(elem);
      }
    }
  }

  // / Añadido de un documento
  public synchronized boolean agregarDocumento(
      DocumentPresentationModel nuevoDoc) throws TransformationException {

    nuevoDoc.cambioAutoImportacion(bActAutomatica.get());
    File ficheroOrigen = nuevoDoc.getFicheroOrigen();
    try {
      docsRecientes.agregar(ficheroOrigen.getCanonicalPath());
    } catch (IOException e) {
      throw new TransformationException(e);
    }

    if (docsAbiertos.containsKey(ficheroOrigen)) {
      docsAbiertos.get(ficheroOrigen).setModeloDocumento(
          nuevoDoc.getModeloDocumento());
      nuevoDoc.cerrar();
      return false;
    } else {
      docsAbiertos.put(ficheroOrigen, nuevoDoc);
      return true;
    }
  }

  // / Cierre de un documento por su nombre
  public synchronized void cierreDocumento(File fCerrado) {
    for (File f : docsAbiertos.keySet()) {
      if (f.equals(fCerrado)) {
        cierreDocumento(docsAbiertos.get(f));
        return;
      }
    }
  }

  // / Cierre del documento actual
  public synchronized void cierreDocumento(
      DocumentPresentationModel docCerrado) {
    // Interrumpimos el vigilado del fichero
    docCerrado.getNotCambios().interrumpir();
    // Retiramos de la lista de abiertos
    docsAbiertos.remove(docCerrado.getFicheroOrigen());
  }

  // / Obtención de la lista de ficheros abiertos
  public File[] getDocumentosAbiertos() {
    return docsAbiertos.keySet().toArray(new File[] {});
  }

  // Obtiene el modelo de presentación para un título determinado. Si se
  // proporciona <code>null</code>, se devuelve el modelo de presentación del
  // documento vacío.
  public DocumentPresentationModel getModeloDocumento(File fich) {
    if (docsAbiertos.containsKey(fich)) {
      return docsAbiertos.get(fich);
    }
    else if (fich == null) {
      return docVacio;
    }
    return null;
  }

  // / Cambio de la hoja de preprocesado por defecto para nuevos documentos
  public synchronized void cambioHojaPreprocesado(
      DocumentPresentationModel docActual, String nombreNuevaHoja)
      throws TransformationException, ParsingException {
    // Si es la misma de antes, no hacemos nada
    if (nombreNuevaHoja.equals(docActual.getNombreHojaPreproc()))
      return;

    // Si la hoja existe, cambiamos la hoja por defecto
    PreprocessingStylesheet nuevaHoja = repoHojas
        .getHojaPreprocesado(nombreNuevaHoja);
    if (nuevaHoja != null) {
      hojaPreprocOmision = nuevaHoja;
      gestorPrefs.setScalarPref(
          DefaultPreferences.CLAVE_PREPROCESADO_OMISION,
          nombreNuevaHoja);

      docActual.cambioHojaPreprocesado(nuevaHoja);
    }
  }

  // / Cambio de la hoja de visualización por defecto para nuevos documentos y
  // para el actual
  public synchronized void cambioHojaVisualizacion(
      DocumentPresentationModel docActual, String nombreNuevaHoja)
      throws TransformationException {

    // Si es la misma de antes, no hacemos nada
    if (nombreNuevaHoja.equals(docActual.getNombreHojaVista()))
      return;

    // Si la hoja existe, cambiamos la hoja por defecto
    ViewStylesheet nuevaHoja = repoHojas
        .getHojaVisualizacion(nombreNuevaHoja);
    if (nuevaHoja != null) {
      hojaVistaOmision = nuevaHoja;
      gestorPrefs.setScalarPref(DefaultPreferences.CLAVE_VISTA_OMISION,
                                nombreNuevaHoja);

      docActual.cambioHojaVisualizacion(nuevaHoja);
    }
  }

  // / Cambio de la autoimportación
  public synchronized void cambioAutoImportacion() {
    // Cambiamos de valor la actualizacion automatica
    final boolean nuevoValor = !bActAutomatica.get();
    fireSetAutoImportacion(nuevoValor);
  }

  private void fireSetAutoImportacion(final boolean nuevoValor) {
    bActAutomatica.set(nuevoValor);

    // Avisamos a los modelos de documento
    for (DocumentPresentationModel doc : docsAbiertos.values()) {
      doc.cambioAutoImportacion(nuevoValor);
    }
  }

  /* --- METODOS DE PAQUETE --- */
  ValueWrapper<String> inicioTarea(String msg) {
    ValueWrapper<String> elem = new ValueWrapper<String>(msg);
    mensajesTareas.offer(elem);
    return elem;
  }

  ArrayList<String> getOrdenEditor(File f) {
    return formatRepository.getFormatoPorFich(f).getOrdenEdicion(f);
  }

  void finTarea(ValueWrapper<String> elem) {
    mensajesTareas.remove(elem);
  }

  /** CLASES INTERNAS * */
  private class ProgramStatusLineFilter implements LineFilter {

    private ValueWrapper<String> previousLine;

    public void processLine(String line) throws Exception {
      synchronized (mensajesTareas) {
        if (previousLine != null) {
          finTarea(previousLine);
        }
        previousLine = inicioTarea(line);
      }
    }

    public void eofReached() throws Exception {
      finTarea(previousLine);
    }
  }

  /**
   * Devuelve una lista con la descripción y extensiones de cada formato
   * disponible.
   *
   * Útil para un selector de ficheros, por ejemplo.
   */
  public List<Pair<String, String[]>> getElemsSelectorFicheros() {
    List<Pair<String, String[]>> lista = new ArrayList<Pair<String, String[]>>();

    for (int i = 0; i < formatRepository.getSize(); ++i) {
      final FormatDescriptor formato = (FormatDescriptor) formatRepository
          .getElementAt(i);
      lista.add(new Pair<String, String[]>(formato.getNombre(), formato
                                           .getExtensiones()));
    }

    return lista;
  }

}
